package Net::Async::AMQP::Server::Consumer;
$Net::Async::AMQP::Server::Consumer::VERSION = '2.000';
use strict;
use warnings;

=head1 NAME

Net::Async::AMQP::Server::Consumer

=head1 VERSION

version 2.000

=head1 SYNOPSIS

=head1 DESCRIPTION

=cut

1;

__END__

=head1 AUTHOR

Tom Molesworth <TEAM@cpan.org>

=head1 LICENSE

Licensed under the same terms as Perl itself, with additional licensing
terms for the MQ spec to be found in C<share/amqp0-9-1.extended.xml>
('a worldwide, perpetual, royalty-free, nontransferable, nonexclusive
license to (i) copy, display, distribute and implement the Advanced
Messaging Queue Protocol ("AMQP") Specification').
